/*
 Silicom SPE net_device driver for Linux.
 Written by Igor Royzis (c) 1998
 Created: December 9, 1998
 Updates: April 14, 1999 bug fixes thanks to Javier Ripoll (jripoll@csi.com)
          March 14, 1999 implemented my_delay instead of SLOW_DOWN_IO;
                         removed stats.tx_bytes in spe_transmit_interrupt
		         (unreferenced there in kernels 2.1+!);

 SPE and relatives low-level procedures

 The external references are:
  1. SPE local structure:
      struct net_device_stats stats;
      unsigned short rx_first;                    // Rx ring buffer start
      unsigned short rx_last;                     // Rx ring buffer end
      unsigned short rx_head;                     // Rx ring buffer head
      unsigned short tx_first;                    // Tx ring buffer start
      unsigned short tx_last;                     // Tx ring buffer end
      unsigned short tx_head;                     // Tx ring buffer head
      unsigned short tx_tail;                     // Tx ring buffer tail
      unsigned short tx_prev;                     // Tx pointer to last sent buffer
      unsigned short tx_prevsize;                 // Last sent buf size (used for chaining)
      unsigned short tx_count;                    // Tx sends active
      unsigned char  curTxCmd;                    // Current Tx command (Xmit, resume xmit)
      unsigned short mem_size;                    // Memory size
      unsigned char  cfgctrl1;                    // Config data1 (used to set/reset promisc)
      unsigned char  cfgctrl2;                    // Config data2 (user to set/reset promisc)
      unsigned short cfgtable[C593TableLength];   // 593 configuration table
      int            cmc_complete;                // Multicast setup indication flag

  2. DEBUG macro. First argument is the debug level. Important debug events
     should be at level 1. Other key events are at level 3 and 4. Level 5
     is very noisy, it is suggested for functions and key points entry-exit
     tracing.
  3. assigning functions addresses for open, close, xmit, interrupt,
     statistics, mcast and reset

*/

#include <pcmcia/config.h>
#include <pcmcia/k_compat.h>

#include <linux/netdevice.h>
#include <linux/etherdevice.h>
#include <linux/skbuff.h>
#include <linux/if_arp.h>
#include <linux/ioport.h>
#include <linux/wanpipe.h>

#include <pcmcia/cs_types.h>
#include <pcmcia/cs.h>
#include <pcmcia/cistpl.h>
#include <pcmcia/ds.h>
#include <sys/io.h>

const char *silspe_version = "silspe.c: v1.3 1999/04/14 Silicom\n";

#include "silspe.h"
#include "speloc.h"

#undef SYNC_XMIT                       // Define for asyncronious xmit
// #define SYNC_XMIT 1
#undef XMIT_DROP
// #define XMIT_DROP 1

/*
*/
#define DSet0()  outb(CSWP0, ioaddr)
#define DSet1()  outb(CSWP1, ioaddr)
#define DSet2()  outb(CSWP2, ioaddr)

#define DisableBoardInts()  outb((inb(ioaddr+B0Mask) | DB0MaskInt), ioaddr+B0Mask)
#ifndef SYNC_XMIT
# define EnableBoardInts()   outb((inb(ioaddr+B0Mask) & ~(DB0MaskRx | DB0MaskTx | DB0MaskExec)), ioaddr+B0Mask)
#else
# define EnableBoardInts()   outb((inb(ioaddr+B0Mask) & ~(DB0MaskRx | DB0MaskExec)), ioaddr+B0Mask)
#endif


/*
 Local function prototypes
*/
static void my_delay(void);

/*
 593 Configuration parameters pattern
*/
static unsigned short ConfigParams[C593TableLength] = {
        02,        // Config Command.
        0,0,       // Not Used.
        0x0C,      // Number of bytes in this command.
        0x002E,    // LoopBack, Preamble, NoSAIns, Address Len; high:Backoff,Priority
        0x0060,    // Interframe spacing; high:Slot Time
        0x08F2,    // Retrys, slot time high, High: CRC Control, BC_DIS, Promiscuous
        0x4000,    // Collision Filter, Carrier Sense; High: Min.frame lenthg
        0x00B6,    // Length/Type.
        0x063f     // DIS_BOF, Multi IA; high: IPG Retrig
};

static int  hardware_send_packet(struct net_device *dev, void *buf, short length);
static void spe_rx(struct net_device *dev);
static void spe_transmit_interrupt(struct net_device *dev);


/*
 Test the 595 memory and fills the memory size. Internal function. Called by
 spe_drvinit. Returns 0 if ok (lp->mem_size has the memory size) or -1 if
 too little memory are detected.
*/
static unsigned short spe_test_memory(struct net_device *dev)
{
unsigned short ioaddr = dev->base_addr;
struct spe_local *lp = (struct spe_local *)dev->priv;
unsigned short memlen = 0, pattern, i;
int result;

    DSet1();
    // Set the limit registers with fake values to do a memory test without
    // erroneous wraps
    outb(0, ioaddr+B1TxLowLim);
    outb(0x7f, ioaddr+B1TxUpLim);
    outb(0x80, ioaddr+B1RxLowLim);
    outb(0xff, ioaddr+B1RxUpLim);
    DSet0();
    // Write a 4k chunk of memory to a rotation pattern. We just write to the
    // first 0x10 words of a 4k
    while(1)  {
       outw(memlen, ioaddr+B0HostAddr);
       for(i=0,pattern=3; i<0x10; i++) {
          outw(pattern, ioaddr+B0MemPort);
          pattern+=3;
       }
       // Read it back
       outw(memlen, ioaddr+B0HostAddr);
       for(i=0,pattern=3; i<0x10; i++)  {
          if(inw(ioaddr+B0MemPort) != pattern)
            break;                        // No more memory
          pattern+=3;
       }

       if(i < 0x10)  {                    // Test failed, no more memory
           result = -1;
           break;
       }

       if(memlen != 0)  {                  // It wasn't first 4k, make sure 1st 4k are zeros
          outw(0, ioaddr+B0HostAddr);
          for(i=0; i<0x10; i++)  {
             if(inw(ioaddr+B0MemPort) != 0)
               break;                      // No more memory
          }
          if(i < 0x10) {                   // Test failed
             result = -1;                  // No more memory
             break;
          }
       }

       memlen += 0x1000;                   // Add 4k to memory size
       // Write the first 4k chunk of memory to zeros. We do this because later this
       // first 4k chunk could get overwritten if we run out of memory
       outw(0, ioaddr+B0HostAddr);
       for(i=0; i<0x10; i++)
          outw(0, ioaddr+B0MemPort);
       if(memlen == 0)  {
         result = 0;
         break;
       }
    } //while

    if(result == -1)  {                 // No more memory
       if(memlen >= 0x2000)  {          // More than 8k memory, good
           lp->mem_size = memlen;
           result = 0;
       }
    }
    else  {                            // We reached max memory in test
       if(memlen >= 0x8000)            // Greater than 32k
          memlen = (((memlen-0x8000) >> 10) + 0x20) | 0x8000;
       if(memlen == 0)                 // Do we have 64k?
          memlen = 0x8040;             // Yes, 64K onboard
       lp->mem_size = memlen;
    }
    return result;
}


/*
 The function initializes the SPE local structure. Must be called before any other
 function in the module.
 Returns: 0 - ok
        ENODEV - 595 chip is not detected
*/
int spe_drvinit(struct net_device *dev)
{
unsigned short ioaddr = dev->base_addr;
struct spe_local *lp = (struct spe_local *)dev->priv;
int i;

    DEBUG(5, "%s: entering drvinit\n", dev->name)
    // Ensure it's 595 chip
    DSet0();
    if ((i=(inb(ioaddr + B0Identity) & IdRegMask)) != IdRegSign)  {
      DEBUG(1, "silspe: Wrong chip id. Read %x instead of %x\n",i, IdRegSign)
      return ENODEV;
    }
    lp->cfgctrl1 = B21_IgnoreBadFrames + B21_TxChainIntMode;
    lp->cfgctrl2 = 0;
    if(0 != spe_test_memory(dev))
      return ENODEV;
    DEBUG(4, "%s: %x memory detected\n", dev->name, lp->mem_size)
    // Copy 593 configuration table
    for(i=0; i < C593TableLength; i++)
       lp->cfgtable[i] = ConfigParams[i];

    DEBUG(5, "%s: exiting drvinit\n", dev->name)
    return 0;
}

/*
 Resets the 595, initializes it and turns to a working state. spe_drvinit
 must be called before. Returns always zero.
*/
int spe_reset(struct net_device *dev)
{
unsigned short ioaddr = dev->base_addr;
int i;
struct spe_local *lp = (struct spe_local *)dev->priv;

    // Reset 595 chip
    DSet0();
    DisableBoardInts();
    outb(CSELR, ioaddr+BCommand);
    for(i=0; i < 10; i++)
      my_delay();
    DSet0();
    DSet2();
    // Ignore bad frames
    outb(inb(ioaddr+B2Control1) | B21_IgnoreBadFrames | lp->cfgctrl1, ioaddr+B2Control1);
    // Clear promiscuous and disable source address insertion
    //outb((inb(ioaddr+B2Control2) & ((~B22_Promiscuous) | lp->cfgctrl2 | B22_NoSrcAddrIns)), ioaddr+B2Control2);
    outb(((inb(ioaddr+B2Control2) & (~B22_Promiscuous)) | lp->cfgctrl2 | B22_NoSrcAddrIns), ioaddr+B2Control2);
    outb((inb(ioaddr+B2Config) & (~B23_TestBits)), ioaddr+B2Config); // Clear test bits
    // Set station address
    for(i=0; i < 6; i++)
      outb(dev->dev_addr[i] , (ioaddr + B2IAddress0 + i));
    /*
     Buffers initialization
    */
    lp->tx_first = TRANSMIT_BUFFER_START;
    lp->tx_head = TRANSMIT_BUFFER_START;
    lp->tx_tail = TRANSMIT_BUFFER_START;
    // Set Transmit Lower limit
    DSet1();
    outb(TRANSMIT_BUFFER_START >> 8, ioaddr+B1TxLowLim);
    // We do not use early transmits
    lp->tx_last = (TRANSMIT_BUFFER_START+TRANSMIT_BUFFER_SIZE-1) & 0xfff8;
#ifndef SYNC_XMIT
    lp->tx_prev = lp->tx_last;
    lp->tx_prevsize = 0;
#endif
    // Set Transmit upper limit
    outb(lp->tx_last >> 8, ioaddr+B1TxUpLim);
    // Set Receive upper and lover limit
    lp->rx_first = (((lp->tx_last >> 8) + 1) << 8) & 0xff00;
    lp->rx_head = lp->rx_first;
    outb(lp->rx_first >> 8, ioaddr+B1RxLowLim);
    outb(lp->mem_size >> 8, ioaddr+B1RxUpLim);
    // Set the Receive stop register to the end of memory
    DSet0();
    lp->rx_last = lp->mem_size;
    outw(lp->rx_last, ioaddr+B0RxStop);

    lp->curTxCmd = CTx;                // Start transmit command
    lp->tx_count = 0;
    // Wait till Execution unit is idle
    DSet0();
    for(i=0; i < 20000; i++)
      if(inb(ioaddr+B0IntStatus) & DB0IntExState)
         break;
    // Configure 595 by read and writing back the config register
    DSet2();
    outb(inb(ioaddr+B2Config), ioaddr+B2Config);
    DSet0();
    // Wait till configuration is complete
    for(i=0; i < 20000; i++)
      if(inb(ioaddr+B0IntStatus) & DB0IntExState)
    // Do selective reset
    outb(CSELR, ioaddr+BCommand);
    // Wait till selective reset is complete
    for(i=0; i < 20000; i++)
      if(inb(ioaddr+B0IntStatus) & DB0IntExState)
         break;
    /*
      Configure 593
    */
    outw(lp->tx_head, ioaddr+B0TxAddr);          // Upadate BAR to the buffer
    outw(lp->tx_head, ioaddr+B0HostAddr);        // Setup memory pointer
    for(i=0; i < C593TableLength; i++)  {
       outw(lp->cfgtable[i], ioaddr+B0MemPort);
    }
    // Issue the config command
    outb(CMConfig, ioaddr+BCommand);
    // Wait till config is complete
    for(i=0; i < 20000; i++)
      if(inb(ioaddr+B0IntStatus) & DB0IntExState)
         break;
    // Acknowledge the Int.
    outb(DB0IntRx+DB0IntTx+DB0IntRxStp+DB0IntExec, ioaddr+B0IntStatus);


    // Set the Transmit BAR to the first Transmit buffer
    outw(lp->tx_head, ioaddr+B0TxAddr);
    // Set the Receive BAR address to the first Receive buffer
    outw(lp->rx_head, ioaddr+B0RxAddr);
    // Set the 595 interrupt
    DSet1();
    outb(inb(ioaddr+B1IntSel) & (~7), ioaddr+B1IntSel);
    // Enable hardware interrupts (not TriState)
    outb(inb(ioaddr+B1Control) | B11_TurnOnInterrupts, ioaddr+B1Control);
    // Start the receive unit and acknowledge any outstanding interrupts
    DSet0();
    outb(CRxENB, ioaddr+BCommand);
    outb(DB0MaskInt, ioaddr+B0IntStatus);
    DSet0();
    // Turn on board interrupts
    EnableBoardInts();
    return 0;
}




/*
*/
int spe_open(struct net_device *dev)
{
struct spe_local *lp = (struct spe_local *)dev->priv;

    DEBUG(5, "spe: entering open routine.\n")

    //dev->tbusy = 0;
    start_net_queue(dev);
    netif_start_queue(dev);
    netif_mark_up(dev);
    spe_reset(dev);
    //dev->interrupt = 0;
    lp->interrupt = 0;
    //dev->start = 1;
    lp->start = 1;
    MOD_INC_USE_COUNT;

    DEBUG(5, "spe: exiting open routine.\n")
    return 0;
}

/*
*/
int spe_start_xmit(struct sk_buff *skb, struct net_device *dev)
{
struct spe_local *lp = (struct spe_local *)dev->priv;
//int ioaddr = dev->base_addr;
int i=0;

    DEBUG(5, "%s: entering spe_send_packet routine.\n", dev->name)
    
    if (!lp->start) return 1;
	
    if (is_queue_stopped(dev)) {
        /* If we get here, some higher level has decided we are broken.
           There should really be a "kick me" function call instead. */
        int tickssofar = jiffies - dev->trans_start;
        if (tickssofar < 40)
            return 1;
        DEBUG(1, "%s: transmit timed out, network cable problem? tx_count=%d\n", dev->name, lp->tx_count)
        lp->stats.tx_errors++;

        /* Try to restart the adaptor. */
        spe_reset(dev);

        //dev->tbusy=0;
	start_net_queue(dev);
        dev->trans_start = jiffies;

    }

    /* If some higher layer thinks we've missed an tx-done interrupt
       we are passed NULL. Caution: dev_tint() handles the cli()/sti()
       itself. */
#if (LINUX_VERSION_CODE < VERSION(2,1,25))
    if (skb == NULL) {
        DEBUG(4, "%s: dev_tint\n", dev->name)
        dev_tint(dev);
        return 0;
    }
    if(skb->len <= 0)
      return 0;
#endif

    /* Block a timer-based transmit from overlapping. */
    //if (test_and_set_bit(0, (void*)&dev->tbusy) != 0)  {
    if (is_queue_stopped(dev) != 0)  {
	stop_net_queue(dev);
        DEBUG(1, "%s: Transmitter access conflict.\n", dev->name)
    }
    else {
        short length = ETH_ZLEN < skb->len ? skb->len : ETH_ZLEN;
        unsigned char *buf = skb->data;
	stop_net_queue(dev);
        i = hardware_send_packet(dev, buf, length);
        dev->trans_start = jiffies;
    }

#ifdef XMIT_DROP
    i=0;
    DEV_KFREE_SKB(skb);
#else
    if (i == 0) {
        DEV_KFREE_SKB(skb);
    }
#endif

    DEBUG(5, "spe: exiting spe_send_packet routine.\n")

    return i;
}

/*
 Handle the network interface interrupts.
*/
void spe_interrupt(int irq, void *dev_id, struct pt_regs * regs)
{
struct net_device *dev = (struct net_device *)DEV_ID;
struct spe_local *lp = (struct spe_local *)dev->priv;
int ioaddr, status;

    DEBUG(5, "spe: entering spe_interrupt routine\n")

    if (dev == NULL) {
        DEBUG (1, "spe_interrupt(): irq %d for unknown net_device\n", irq)
        return;
    }
    if (!lp->start) {
        DEBUG (1, "spe_interrupt(): device stopped\n", irq)
        return;
    }


    //dbg
    //if(dev->interrupt == 1)
    //if(lp->interrupt == 1)
    //  DEBUG(1,"spe: nested interrupts!\n")
    //dbg
	
    if (!netif_device_present(dev)) return;

    //dev->interrupt = 1;
    lp->interrupt = 1;
    ioaddr = dev->base_addr;

    DisableBoardInts();
    DSet0();
    status = inb(ioaddr + B0IntStatus);

    while(status & (DB0IntRx+DB0IntTx+DB0IntRxStp+DB0IntExec)) {
        if((status & 0xff) == 0xff)              // False interrupt if card removed
           break;
        outb(status, ioaddr+B0IntStatus);        // Acknowledge interrupt
        if (status & DB0IntRx) {
            DEBUG(4, "spe: packet received interrupt\n")
            spe_rx(dev);                         // Receive packets
        }

        if (status & DB0IntTx) {
            DEBUG(4, "spe: packet transmit interrupt\n")
            spe_transmit_interrupt(dev);         // Process the status of transmitted packets
        }
        if (status & DB0IntExec) {
            DEBUG(4, "spe: exec interrupt\n")
            if((inb(ioaddr+BCommand) & 0x1f) == CMCSET)  {  // Multicast completion
                struct spe_local *lp = (struct spe_local *)dev->priv;
                outw(lp->tx_head, ioaddr+B0TxAddr);         // Update transmit BAR/CAR
                lp->curTxCmd= CTx;                          // Next transmit command
                lp->cmc_complete = 1;                       // Indication mcast setup completed
            }
        }
        status = inb(ioaddr + B0IntStatus);
    } // while
    //dev->interrupt = 0;
    lp->interrupt = 0;
    EnableBoardInts();
    DEBUG(5, "spe: exiting spe_interrupt routine\n")
    return;
}

/*
*/
int spe_close(struct net_device *dev)
{
struct spe_local *lp = (struct spe_local *)dev->priv;
unsigned short ioaddr = dev->base_addr, i;

    //dev->tbusy = 1;
    stop_net_queue(dev);
    netif_mark_down(dev);
    //dev->start = 0;
    lp->start = 0;

    DSet1();
    outb((inb(ioaddr+B1Control) & 0x7f), ioaddr+B1Control);  // Disable interrupt line
    DSet0();
    outb(CSTPRx, ioaddr+BCommand);               // Stop RX
    lp->tx_first = lp->tx_head = lp->tx_tail = TRANSMIT_BUFFER_START ;

    DisableBoardInts();                          // Mask all interrupts
    outb(DB0MaskInt, ioaddr+B0IntStatus);        // Clear outstanding interrupts
    outb(CSELR, ioaddr+BCommand);                // Reset the chip
    for(i=0; i < 10; i++)
       my_delay();

    MOD_DEC_USE_COUNT;
    return 0;
}

/*
 Get the current statistics.  This may be called with the card open or
 closed.
*/
struct net_device_stats *spe_get_stats(struct net_device *dev)
{
struct spe_local *lp = (struct spe_local *)dev->priv;

    return &lp->stats;
}

/*
 Set or clear the multicast filter / mode
*/
void set_multicast_list(struct net_device *dev)
{
struct spe_local *lp = (struct spe_local *)dev->priv;
short ioaddr = dev->base_addr;
struct dev_mc_list *dmi=dev->mc_list;

    if (dev->flags&(IFF_ALLMULTI|IFF_PROMISC) || dev->mc_count > 63) {
        dev->flags|=IFF_PROMISC;
        /* Enter the promiscuous mode */
//        lp->cfgctrl1 &= ~B21_IgnoreBadFrames;    // Promiscuous mode receives bad frames
        lp->cfgctrl2 = B22_Promiscuous;            // Set promiscuous mode
        lp->cfgtable[C593ModeIndex]  |= C593_EnablePromisc;
        spe_reset(dev);                            // Reset the net_device
        DEBUG(1, "%s: promiscuous mode enabled.\n", dev->name)
    }
    else if (dev->mc_count == 0 )  {
        /* Disable the promiscuous mode and multicasts */
        lp->cfgctrl1 |= B21_IgnoreBadFrames;    // Discard bad frames
        lp->cfgctrl2 = 0;                       // Unset promiscuous mode
        lp->cfgtable[C593ModeIndex]  &= ~C593_EnablePromisc;
        spe_reset(dev);                         // Reset the net_device
    }
    else  {
        /* Write down the multicast table. the multicast table is in dev->mc_list(dmi),
           the number of addresses in the table is dev->mc_count
        */
        unsigned short *eaddrs, i;
        lp->cmc_complete = 0;                    // Clear the indication
        // Wait until EXE-DMA is idle before setting the BAR
        while(inb(ioaddr+B0IntStatus) & DB0IntExState)
           ;
        DisableBoardInts();
        // Setup 595 header for multicast
        outw(lp->tx_head, ioaddr+B0TxAddr);
        outw(lp->tx_head, ioaddr+B0HostAddr);
        outw(CMCSET, ioaddr+B0MemPort);          // Set multicast command
        outw(0, ioaddr+B0MemPort);               // Zero status
        outw(0, ioaddr+B0MemPort);               // Zero link
        outw(6*dev->mc_count, ioaddr+B0MemPort); // Size of multicast addresses
        for (i = 0; i < dev->mc_count; i++)  {   // Write down the mc addresses
            eaddrs=(unsigned short *)dmi->dmi_addr;
            dmi=dmi->next;
            outw(*eaddrs++, ioaddr+B0MemPort);
            outw(*eaddrs++, ioaddr+B0MemPort);
            outw(*eaddrs++, ioaddr+B0MemPort);
        }
        EnableBoardInts();
        outb(CMCSET, ioaddr+BCommand);           // Issue multicast setup command
        // Wait EXE int occur
        for(i=0; (i < 30000) && (lp->cmc_complete == 0); i++)
          ;
        if(lp->cmc_complete == 0)  {
           DEBUG(1, "%s: Multicast setup failed\n", dev->name)
        }
    }
}

/*
 Issue the transmit commad. Called by spe_start_xmit
*/
static int hardware_send_packet(struct net_device *dev, void *buf, short length)
{
struct spe_local *lp = (struct spe_local *)dev->priv;
short ioaddr = dev->base_addr;
unsigned short next, avail;
#define LOOP_DELAY 200


    DEBUG(5, "spe: entering hardware_send_packet routine.\n")

    // Disable interrupts
    DisableBoardInts();

#ifdef SYNC_XMIT
    // Wait until 595 is idle for XMIT command
    // Used for synchronious xmit
    for(avail=0; avail < LOOP_DELAY; avail++)  {
       if((inb(ioaddr+B0IntStatus) & DB0IntExState) == 0)
          break;
       my_delay();
    }
    if(avail == LOOP_DELAY)  {
      DEBUG(1, "spe: transmitter is busy\n")
      dev->tbusy = 1;
      lp->tx_count = 1;
      EnableBoardInts();
      return 1;
    }
#else
    // Check if we have a space in the Tx buffer
    if(lp->tx_tail > lp->tx_head)
       avail = lp->tx_tail - lp->tx_head;
    else
       avail = lp->tx_tail - lp->tx_head + TRANSMIT_BUFFER_SIZE - 8;

    if(avail < (TRANSMIT_FRAME_LENGTH + 10))  {
           DEBUG(1, "spe: No available buffers? head=%x,tail=%x,start=%x,end=%x,count=%d\n",
                    lp->tx_head, lp->tx_tail, lp->tx_first, lp->tx_last, lp->tx_count);
           //dev->tbusy = 1;
	   stop_net_queue(dev);
           EnableBoardInts();
           DEBUG(5, "spe: Exitting hardare_send_packet, net_device is busy\n")
           return 1;
    }
#endif
    outw(lp->tx_head + HEADER_LENGTH, ioaddr+B0HostAddr);  //   Frame start
    outsw(ioaddr+B0MemPort, buf, ((length+1)>>1)); // Copy the frame

    outw(lp->tx_head, ioaddr+B0HostAddr);    // Setup memory pointer
    outw(CTx, ioaddr+B0MemPort);             // Clear done flag and set command
    outw(0, ioaddr+B0MemPort);               // Zero status
#ifndef SYNC_XMIT
    // Set the link to next buffer
    next = (lp->tx_head + length + HEADER_LENGTH + 1) & 0xfffe;
    if(next >= lp->tx_last)  {               // Wrap
       if(next < TRANSMIT_BUFFER_SIZE)
         next = next - TRANSMIT_BUFFER_SIZE + 8; // Correct for boundary wrap
       else
         next -=  TRANSMIT_BUFFER_SIZE;
     }
#else
     next = 0;                                // No next on synch. xmit
#endif
     outw(next, ioaddr+B0MemPort);
     outw(length, ioaddr+B0MemPort);          // Frame length
#ifndef SYNC_XMIT
     outw(lp->tx_prev+6, ioaddr+B0HostAddr);  // Previous Tx, offset to byte count field
     outw(lp->tx_prevsize | TxChainBit, ioaddr+B0MemPort); // Previous tx size+chain bit
     lp->tx_prevsize = length;
     lp->tx_prev = lp->tx_head;
     lp->tx_head = next;
#else
      outw(0, ioaddr+B0TxAddr);
#endif
     outb(lp->curTxCmd, ioaddr+BCommand);      // Start transmit
#ifndef SYNC_XMIT
     lp->curTxCmd = CRSTx;                     // Next time it will be restart
     lp->tx_count++;
#else
     lp->stats.tx_packets++;
#if (LINUX_VERSION_CODE >= VERSION(2,1,25))
     lp->stats.tx_bytes += length;
#endif
#endif
//     dev->tbusy = 0;
//     mark_bh(NET_BH);                         // Inform upper layer
     EnableBoardInts();
     DEBUG(5, "spe: Exitting hardware_send_packet\n")
     return 0;
}


/*
*/
static void spe_rx(struct net_device *dev)
{
struct spe_local *lp = (struct spe_local *)dev->priv;
short ioaddr = dev->base_addr;
unsigned short rcv_event, rcv_status, rcv_size, next, temp;
static unsigned short temp2;
static int maxrxf = 0;
int rxf=0;

    DEBUG(5, "spe: entering spe_rx routine.\n")

    while(1)  {
      rxf++;
      if (rxf > maxrxf) maxrxf = rxf;
      outw(temp=lp->rx_head, ioaddr+B0HostAddr);  // Set the pointer to start of RCV
      rcv_event = inw(ioaddr+B0MemPort);          // Get the event
      if(!(rcv_event&RxDone))
         break;                                   // Nothing done yet
      rcv_status = inw(ioaddr+B0MemPort);         // Read the status
      if(rcv_status &  (RXS_ReceiveOK))  {
         next = inw(ioaddr+B0MemPort);            // Read the next link
         rcv_size = inw(ioaddr+B0MemPort);        // Read the size
         lp->rx_head = next;           		  // Next frame
         next -= 2;                    		  // Next stop
         if(next < lp->rx_first)
           outw(temp2=lp->rx_last, ioaddr+B0RxStop);
         else
           outw(temp2=next, ioaddr+B0RxStop);
         if((rcv_size >= MIN_PACKET) && (rcv_size <= MAX_PACKET)) {  // for valid packets only
            struct sk_buff *skb;
            skb = ALLOC_SKB(rcv_size+3);
            if (skb == NULL) {
                DEBUG(1, "%s: No buffers available, dropping packet.\n", dev->name)
                lp->stats.rx_dropped++;
                break;
            }
            skb->dev = dev;
#define BLOCK_INPUT(buf,len) insw(ioaddr+B0MemPort, buf, (len + 1) >> 1)
            GET_PACKET(dev, skb, rcv_size);
            lp->stats.rx_packets++;
#if (LINUX_VERSION_CODE >= VERSION(2,1,25))
            lp->stats.rx_bytes += skb->len;
#endif
            netif_rx(skb);
         }
         else {
            DEBUG(1, "spe: bad frame length received. status: %x, length: %d\n", rcv_status, rcv_size)
         }
         // Discard the packet
      } //if
      else  {
        // Bad frame is here. Discard? reset?
         lp->stats.rx_errors++;
         lp->stats.rx_packets++;
         next = inw(ioaddr+B0MemPort);            // Read the next link
         DEBUG(1, "spe: bad frame received status: %x i=%ld h=%x n=%x na=%x rx=%d mrx=%d\n", rcv_status,lp->stats.rx_packets, temp,next,temp2,rxf,maxrxf)
         rcv_size = inw(ioaddr+B0MemPort);        // Read the size
         // Discard the packet
         lp->rx_head = next;           		  // Next frame
         next -= 2;                    		  // Next stop
         if(next < lp->rx_first)
           outw(temp2=lp->rx_last, ioaddr+B0RxStop);
         else
           outw(temp2=next, ioaddr+B0RxStop);
         // break;
      }
    } //while
    DEBUG(5, "spe: exitting spe_rx routine\n")
}


/*
*/
static void spe_transmit_interrupt(struct net_device *dev)
{
struct spe_local *lp = (struct spe_local *)dev->priv;
unsigned short ioaddr = dev->base_addr, next;
unsigned short xmt_status, avail;

    if (!lp->start) return;

    while (lp->tx_count > 0) {
        outw(lp->tx_tail, ioaddr+B0HostAddr);    // Set the address of buffer
        if(!(inw(ioaddr+B0MemPort) & TxDone))
          break;                                 // Buffer is not yet done
        xmt_status = inw(ioaddr+B0MemPort);      // Read the status
        next = inw(ioaddr+B0MemPort);            // Read the link
        if(xmt_status & TXS_TransmitOK)  {
          lp->stats.tx_packets++;
        }
        else if(xmt_status & TXS_TransmitFail)  { // If a fatal failure
            DEBUG(1,"spe: transmit failure status: %x\n", xmt_status)
            lp->stats.tx_errors++;
           // Skip the Tx failed and try the next if available
           outw(next, ioaddr+B0TxAddr);          // Update transmit BAR/CAR
           if(lp->tx_count != 1)
             outb(CTx, ioaddr+BCommand);         // Restart transmit
           else
             lp->curTxCmd = CTx;                 // Reset to Xmit command
        }
        lp->tx_count--;
#ifndef SYNC_XMIT
        lp->tx_tail = next;
#else
        DEBUG(4,"spe: transmit OK\n")
#endif
    } //while
    if (is_queue_stopped(dev) != 0) {
        // Check if we have a space in the Tx buffer
        if(lp->tx_tail > lp->tx_head)
           avail = lp->tx_tail - lp->tx_head;
        else
           avail = lp->tx_tail - lp->tx_head + TRANSMIT_BUFFER_SIZE - 8;

        if(avail >= (TRANSMIT_FRAME_LENGTH + 10))  {
           //dev->tbusy = 0;
           //mark_bh(NET_BH);                         // Inform upper layer
	   wake_net_dev(dev);
        }
    }
}

/*
 SLOW_DOWN_IO macro has been removed from 2.2 kernel (actually changed the
 name) so I implemented a simple delay function right here.
 Check out if safe for other platforms..
*/
static void my_delay()
{
    inb(0x61);
    inb(0x61);
    inb(0x61);
}
